/* This director extends FSMDirector by consuming only input tokens
   that are needed in the current state.

   Copyright (c) 2004-2005 The Regents of the University of California.
   All rights reserved.
   Permission is hereby granted, without written agreement and without
   license or royalty fees, to use, copy, modify, and distribute this
   software and its documentation for any purpose, provided that the above
   copyright notice and the following two paragraphs appear in all copies
   of this software.

   IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
   FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
   ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
   THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.

   THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
   PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
   CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
   ENHANCEMENTS, OR MODIFICATIONS.

   PT_COPYRIGHT_VERSION_2
   COPYRIGHTENDKEY
*/
package ptolemy.domains.fsm.kernel;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ptolemy.actor.CompositeActor;
import ptolemy.actor.IOPort;
import ptolemy.actor.Receiver;
import ptolemy.data.expr.ASTPtAssignmentNode;
import ptolemy.data.expr.ASTPtRootNode;
import ptolemy.data.expr.ParseTreeFreeVariableCollector;
import ptolemy.data.expr.ParserScope;
import ptolemy.data.expr.PtParser;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;


//////////////////////////////////////////////////////////////////////////
//// NonStrictFSMDirector

/**
   This director extends FSMDirector by consuming only input tokens that
   are needed in the current state. An input port will consume one
   token if the port name appears in: (1) at least one guard expression in any
   transition leaving the current state and/or (2) the output actions
   of the enabled transition.

   FIXME: This is highly preliminary. Missing capabilities:
   FIXME: The input ports appeared in SetActions of the enabled transition
   should also be included.
   FIXME: Doesn't handle multi-ports.
   FIXME: Currently this director uses the default receiver of FSMDirector,
   which is a mailbox, so there is no way to consume more than one token.
   This director could use a different receiver and support a syntax in
   the guard expression language to support consumption of more than one
   token.
   FIXME: The current implementation also does not support state
   refinement. That is, the refinement is not examined to determine
   whether to consume a token.

   @author Rachel Zhou
   @version $Id: NonStrictFSMDirector.java,v 1.7 2005/03/01 04:48:27 zhouye Exp $
   @since Ptolemy II 4.1
   @Pt.ProposedRating Red (zhouye)
   @Pt.AcceptedRating Red (cxh)
*/
public class NonStrictFSMDirector extends FSMDirector {
    /** Construct a director in the default workspace with an empty string
     *  as its name. The director is added to the list of objects in
     *  the workspace. Increment the version number of the workspace.
     */
    public NonStrictFSMDirector() {
        super();
    }

    /** Construct a director in the given container with the given name.
     *  The container argument must not be null, or a
     *  NullPointerException will be thrown.
     *  If the name argument is null, then the name is set to the
     *  empty string. Increment the version number of the workspace.
     *  @param container Container of this director.
     *  @param name Name of this director.
     *  @exception IllegalActionException If the name has a period in it, or
     *   the director is not compatible with the specified container.
     *  @exception NameDuplicationException If the container not a
     *   CompositeActor and the name collides with an entity in the container.
     */
    public NonStrictFSMDirector(CompositeEntity container, String name)
            throws IllegalActionException, NameDuplicationException {
        super(container, name);
    }

    /** Construct a director in the  workspace with an empty name.
     *  The director is added to the list of objects in the workspace.
     *  Increment the version number of the workspace.
     *  @param workspace The workspace of this director.
     */
    public NonStrictFSMDirector(Workspace workspace) {
        super(workspace);
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Get the enabled transition and the referred input ports in its
     *  outputActions. If they are not referred by the guard expressions,
     *  transfer these inputs before the super class fire() is called.
     *  @exception IllegalActionException If the super class throws it.
     */
    public void fire() throws IllegalActionException {
        FSMActor controller = getController();
        controller._readInputs();

        State currentState = controller.currentState();
        Transition enabledTransition = controller._checkTransition(currentState
                .nonpreemptiveTransitionList());
        getOutputActionsReferredInputPorts(enabledTransition);

        CompositeActor container = (CompositeActor) getContainer();
        List inputPortList = container.inputPortList();
        _consumeToken = true;

        for (int i = 0; i < inputPortList.size(); i++) {
            IOPort port = (IOPort) inputPortList.get(i);

            if (_outputActionReferredInputPorts.contains(port)
                    && !_guardReferredInputPorts.contains(port)) {
                Receiver[][] insideReceivers = (Receiver[][]) port.getReceivers();
                super.transferInputs(port);
                controller._readInputs();
            }
        }

        super.fire();
    }

    /** Given a state, get a list of referred input ports in the guard
     *  expressions of all the transitions that go out from this state.
     * @param currentState The given state.
     * @exception IllegalActionException If there is no controller or if
     *  the guard expression is illegal.
     */
    public void getGuardReferredInputPorts(State currentState)
            throws IllegalActionException {
        _guardReferredInputPorts.clear();

        Iterator transitions = currentState.nonpreemptiveTransitionList()
            .iterator();

        while (transitions.hasNext()) {
            Transition transition = (Transition) transitions.next();
            String string = transition.getGuardExpression();

            if (string == "") {
                throw new IllegalActionException(this,
                        "guard expression on " + transition.getName() + "is null!");
            }

            PtParser parser = new PtParser();
            ASTPtRootNode parseTree = parser.generateParseTree(string);
            ParseTreeFreeVariableCollector variableCollector = new ParseTreeFreeVariableCollector();
            FSMActor controller = getController();
            ParserScope scope = controller.getPortScope();
            Set set = variableCollector.collectFreeVariables(parseTree, scope);
            getReferredInputPorts(set, _guardReferredInputPorts);
        }
    }

    /** Given a transition, get a list of referred input ports in the
     *  outputActions of that transition.
     * @param transition The transition.
     * @exception IllegalActionException If there is no controller or if
     *  the outputActions is illegal.
     */
    public void getOutputActionsReferredInputPorts(Transition transition)
            throws IllegalActionException {
        _outputActionReferredInputPorts.clear();

        String string = transition.outputActions.getExpression();
        PtParser parser = new PtParser();
        ASTPtRootNode parseTree;
        ParseTreeFreeVariableCollector variableCollector = new ParseTreeFreeVariableCollector();
        FSMActor controller = getController();
        ParserScope scope = controller.getPortScope();

        if (string != "") {
            Map map = parser.generateAssignmentMap(string);
            Set set = new HashSet();

            for (Iterator names = map.keySet().iterator(); names.hasNext();) {
                String name = (String) names.next();

                ASTPtAssignmentNode node = (ASTPtAssignmentNode) map.get(name);
                parseTree = node.getExpressionTree();
                set = variableCollector.collectFreeVariables(parseTree, scope);
                getReferredInputPorts(set, _outputActionReferredInputPorts);
            }
        }
    }

    /** Given a set of ports, get those are input ports and put them
     *  in the indicated referred list.
     * @param set The given set of ports
     * @param referredList The referred list.
     */
    public void getReferredInputPorts(Set set, List referredList) {
        CompositeActor container = (CompositeActor) getContainer();
        List inputPortList = container.inputPortList();

        for (int i = 0; i < inputPortList.size(); i++) {
            IOPort inputPort = (IOPort) inputPortList.get(i);

            if (set.contains(inputPort.getName())) {
                referredList.add(inputPort);
            }
        }
    }

    /** Initialize the director. Get the referred input ports
     *  in the guard expressions of all the transitions that go out
     *  from the initial state.
     *  @exception IllegalActionException If the super class throws it.
     */
    public void initialize() throws IllegalActionException {
        super.initialize();
        _consumeToken = false;

        FSMActor controller = getController();

        getGuardReferredInputPorts(controller.getInitialState());
    }

    /** Get the referred input ports in the guard expressions
     *  of all the transitions that go out from the current state.
     *  Then call super class postfire().
     *  @exception IllegalActionException If the super class throws it.
     */
    public boolean postfire() throws IllegalActionException {
        boolean postfireValue = super.postfire();
        FSMActor controller = getController();
        getGuardReferredInputPorts(controller.currentState());
        _consumeToken = false;
        return postfireValue;
    }

    /** Override by super class by only transferring inputs for those
     *  input ports that appear in at least one guard expression
     *  of all the transitions that go out from the current state.
     */
    public boolean transferInputs(IOPort port) throws IllegalActionException {
        if (_guardReferredInputPorts.contains(port)) {
            return super.transferInputs(port);
        } else {
            return true;
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////
    // A flag indicates whether the receiver is a peek one or not.
    private boolean _consumeToken;

    // A list of input ports that appear at least in one guard
    // expression in all the transitions that go from the current state,
    private LinkedList _guardReferredInputPorts = new LinkedList();

    // A list of input ports that appear in the output actions of
    // the enabled transition.
    private LinkedList _outputActionReferredInputPorts = new LinkedList();
}
